/*
 * This file is part of libFirm.
 * Copyright (C) 2012 University of Karlsruhe.
 */

/**
 * @file
 * @brief   declarations for TEMPLATE backend -- private header
 */
#ifndef FIRM_BE_TEMPLATE_TEMPLATE_BEARCH_T_H
#define FIRM_BE_TEMPLATE_TEMPLATE_BEARCH_T_H

#endif
