/*
 * This file is part of libFirm.
 * Copyright (C) 2012 University of Karlsruhe.
 */

/**
 * @file
 * @brief   Node/edge bucket data types.
 * @date    30.11.2008
 * @author  Sebastian Buchwald
 */
#ifndef KAPS_BUCKET_T_H
#define KAPS_BUCKET_T_H

#include "pbqp_t.h"

typedef pbqp_node_t** pbqp_node_bucket_t;
typedef pbqp_edge_t** pbqp_edge_bucket_t;

#endif
