/*
 * This file is part of libFirm.
 * Copyright (C) 2017 University of Karlsruhe.
 */

#ifndef FIRM_BE_MIPS_TRANSFORM_H
#define FIRM_BE_MIPS_TRANSFORM_H

#include "firm_types.h"

void mips_transform_graph(ir_graph *irg);

#endif
