/*
 * This file is part of libFirm.
 * Copyright (C) 2012 University of Karlsruhe.
 */

/**
 * @file
 * @brief   Brute force PBQP solver.
 * @date    02.12.2008
 * @author  Sebastian Buchwald
 */
#ifndef KAPS_BRUTE_FORCE_H
#define KAPS_BRUTE_FORCE_H

#include "pbqp_t.h"

void solve_pbqp_brute_force(pbqp_t *pbqp);

#endif
