/*
 * This file is part of libFirm.
 * Copyright (C) 2012 University of Karlsruhe.
 */

/**
 * @file
 * @author Michael Beck
 * @brief  Reassociation optimization.
 */
#ifndef REASSOC_T_H
#define REASSOC_T_H

void ir_register_reassoc_node_ops(void);

#endif
