/*
 * This file is part of libFirm.
 * Copyright (C) 2015 Matthias Braun
 */

/**
 * @file
 * @brief   nodemap -- private header.
 * @author  Matthias Braun
 */
#ifndef FIRM_IRNODEMAP_T_H
#define FIRM_IRNODEMAP_T_H

typedef struct ir_nodemap {
	void **data;
} ir_nodemap;

#endif
