/*
 * This file is part of libFirm.
 * Copyright (C) 2012 University of Karlsruhe.
 */

/**
 * @file
 * @brief   Heuristic PBQP solver.
 * @date    02.10.2008
 * @author  Sebastian Buchwald
 */
#ifndef KAPS_HEURISTICAL_CO_H
#define KAPS_HEURISTICAL_CO_H

#include "pbqp_t.h"

void solve_pbqp_heuristical(pbqp_t *pbqp);

#endif
