/*
 * This file is part of libFirm.
 * Copyright (C) 2012 Karlsruhe Institute of Technology.
 */
#ifndef HEURISTICAL_CO_LD_H_
#define HEURISTICAL_CO_LD_H_

#include "pbqp_t.h"
#include "deq.h"

void solve_pbqp_heuristical_co_ld(pbqp_t *pbqp, deq_t *rpeo);

#endif
