/*
 * This file is part of libFirm.
 * Copyright (C) 2017 University of Karlsruhe.
 */

#ifndef FIRM_BE_MIPS_MIPS_BEARCH_T_H
#define FIRM_BE_MIPS_MIPS_BEARCH_T_H

#define MIPS_MACHINE_SIZE 32

#endif
