/*
 * This file is part of libFirm.
 * Copyright (C) 2012 University of Karlsruhe.
 */

/**
 * @file
 * @author  Matthias Braun
 */
#ifndef LPP_GUROBI_H
#define LPP_GUROBI_H

#include "lpp.h"

void lpp_solve_gurobi(lpp_t *lpp);

#endif
