/**
 * @file
 * @brief Function prototypes for the emitter functions.
 * @note  DO NOT EDIT THIS FILE, your changes will be lost.
 *         Edit ./ir/be/TEMPLATE/TEMPLATE_spec.pl instead.
 *         created by: ./ir/be/scripts/generate_emitter.pl ./ir/be/TEMPLATE/TEMPLATE_spec.pl build/gen/ir/be/TEMPLATE
 * @date  Tue Jul  7 23:51:18 2026
 */
#ifndef FIRM_BE_TEMPLATE_GEN_TEMPLATE_EMITTER_H
#define FIRM_BE_TEMPLATE_GEN_TEMPLATE_EMITTER_H

void TEMPLATE_register_spec_emitters(void);
void TEMPLATE_register_spec_binary_emitters(void);

#endif
